from snpio.analysis.genotype_encoder import GenotypeEncoder
from snpio.filtering.nremover2 import NRemover2
from snpio.io.phylip_reader import PhylipReader
from snpio.io.structure_reader import StructureReader
from snpio.io.vcf_reader import VCFReader
from snpio.plotting.plotting import Plotting
